#include "tcp_server.h"

static void reaper(int sig_num);  // this one has to be outside of class

static void reaper(int sig_num)
 {
   int   status;

    while (wait3(&status, WNOHANG, (struct rusage *)0) >= 0)
                /* empty */;
    signal(SIGCHLD, reaper);
 }

void TcpServer::start_server(int port, int listen_queue_size, char* log_file)
 {
  if(!log_file)
   log = stderr;
  else if(!strcmp("-", log_file))
    log = stderr;
  else
   {
    log = fopen(log_file, "a");
    if(!log)
     {
      fprintf(stderr, "Tcp Server: could not open log file %s : %s\n", 
        log_file, strerror(errno));
      exit(1);
     }
   }
  
  setbuf(log, NULL);

  struct protoent *tcp_prot = getprotobyname("tcp");
  if(!tcp_prot)
   fatal_error("TcpServer: tcp does not seem to be supported");

  socket_fd = socket(AF_INET,SOCK_STREAM, tcp_prot->p_proto);

  if(socket_fd == -1)
   fatal_error("TcpServer: error creating the socket: %s", strerror(errno));
  
  memset(&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);
  serv_addr.sin_family = AF_INET;
  
  int sock_len = sizeof(serv_addr) ; 
 
  if(bind(socket_fd, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) 
                        == -1)
    {
     fatal_error("TcpServer: Could not bind: %s", strerror(errno));
    }

   if(listen(socket_fd, listen_queue_size) == -1)
    fatal_error("TcpServer: could not listen: %s", strerror(errno));



 }

void TcpServer::run_server()
 {
   
   pid_t fork_pid = fork();

   if(fork_pid == -1)
    fatal_error("TcpServer: could not fork into background: %s", strerror(errno));

   signal(SIGCHLD, reaper);

   if(fork_pid > 0) 
    {
     close(socket_fd);
     return;
    }

   int serv_addr_size = sizeof(serv_addr);

   while(1)
    {
     int client_socket_fd = accept(socket_fd, (struct sockaddr*)&serv_addr, &serv_addr_size);
     
     if(client_socket_fd == -1)
      {
       if(errno != EINTR)
        warn( "TcpServer: error on accept : %s\n", strerror(errno));
      }
     else
      {
       int client_fork_pid = fork();
       if(client_fork_pid == -1)
        warn("TcpServer: could not fork off a client handler : %s\n",
	 strerror(errno));
       else
        if(client_fork_pid == 0)
	 {
	  handle_client(client_socket_fd);
	  //fprintf(stderr, "Closing the client socket\n");
	  if(close(client_socket_fd) == -1); 
	   //warn( "Failed to close the client socked %s\n",
	   // strerror(errno));
	  return;
	 }
	else
	 close(client_socket_fd);
      }
    }
 }
   
TcpServer::TcpServer(int port,int listen_queue_size, char* log_file)
 {
  start_server(port, listen_queue_size, log_file);
 }

TcpServer::~TcpServer()
 {
 }

char* TcpServer::get_peer_addr(int client_socket_fd)
 {
  struct sockaddr_in peername;
  int peername_len = sizeof(sockaddr_in);
  if(getpeername(client_socket_fd, (struct sockaddr*)&peername, &peername_len) == -1)
   warn("Cannot get peer");
  else
   {
    char* src_addr = inet_ntoa(peername.sin_addr);
    return src_addr;
   }
  return "unknown";
 }

void TcpServer::cleanup()
 {
  close(socket_fd);
 }

#ifdef DEBUG

class TestServer: public TcpServer
 {
  protected:

    void handle_client(int client_socket_fd);

  public:

   TestServer(int port, int queue_size, char* log_file) :  
    TcpServer(port, queue_size, log_file)
    {
    }
   
 };



void TestServer::handle_client(int client_socket_fd)
 {
  char buf[512];
  //write(client_socket_fd,"Enter a 5 letter word:", strlen("Enter a 5 letter word:"));
  //read(client_socket_fd, buf, sizeof(buf));
  //write(client_socket_fd,"You have entered : ", strlen("You have entered : "));
  //buf[5] = '\n';
  //write(client_socket_fd, buf, 6);
  
  //ofstream os(client_socket_fd);
  //ifstream is(client_socket_fd);

  //os.setbuf(NULL);
  //is.setbuf(NULL);

  //os << "Please type your name:";
  //is.getline(buf, 512);
  //os << buf << ", good bye" << endl;

  info("connect from %s ", get_peer_addr(client_socket_fd));

  FILE *in  = fdopen(client_socket_fd, "r");
  FILE *out  = fdopen(client_socket_fd, "w");

  setbuf(in, NULL);
  setbuf(out, NULL);

  if(!in)
   {
    warn("fdopen failed on 'in': %s\n", strerror(errno));
    return;
   }

  if(!out)
   {
    warn("fdopen failed on 'out': %s\n", strerror(errno));
    return;
   }

  fprintf(out, "Please enter your name:");
  fgets( buf, sizeof(buf), in);
  fprintf(out, "Good-bye, %s\n", buf);
 }

int main(int c, char** argv)
 {
  int port = 1234;
  char* log_file = "-";
  if(c > 1) port = atoi(argv[1]);
  if(c > 2) log_file = argv[2];

    TestServer s(port, 50, log_file);
    s.run_server();
 }

#endif 
